/*
 * linux/kernel/chr_drv/labcard.c
 * v. 0.0.1 (c) 01/27/1995 9:04:33
 * R. Gosiorovsky & T. Hruz & L. Koren
 */

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/labcard.h>
#include <asm/io.h>
#include <asm/segment.h>

#define LABCARD_MAJOR 19

#define LABCARD_INPUT_CHANNEL 16
#define LABCARD_OUTPUT_CHANNEL 2
#define LABCARD_MAX LABCARD_INPUT_CHANNEL + LABCARD_OUTPUT_CHANNEL
#define LABCARD_MIN 0
#define LABCARD_OUTPUT_CHANNEL_0 16
#define LABCARD_OUTPUT_CHANNEL_1 17

static short     baseI=0x220;              /* default base address */
static short     channel_busy[LABCARD_MAX];

static short read_AD(char channel)
{
 unsigned short adhigh;
 unsigned short adlow;
 unsigned short ad;
 int timcnt=0;

 outb(channel,(unsigned short)(baseI+10));    /* MUX selection    */
 outb((char)0,(unsigned short)baseI+12);      /* software trigger */

 do
  {
    adhigh = inb((unsigned short)(baseI+5));
    if (++timcnt >= 200) return(0x3fff);      /* bad read */
  }
 while ((adhigh&0x10) != 0);

 adlow = inb((unsigned short)(baseI+4));
 ad = (adhigh<<8) + adlow;

 return(ad);
}

static int write_DA(char channel,short data)
{
   switch(channel)
    {
       case 0:
               outb((char)(data&0xff),(unsigned short)(baseI+4));
               outb((char)((data&0x0f00)>>8),(unsigned short)(baseI+5));
	       return 1;
       case 1:
               outb((char)(data&0xff),(unsigned short)(baseI+6));
               outb((char)((data&0x0f00)>>8),(unsigned short)(baseI+7));
	       return 1;
       default:
	       return 0;
     }
}

static int labcard_open(struct inode * inode, struct file * file)
{
	int minor = MINOR(inode->i_rdev);
	if((minor < LABCARD_MIN)||(minor >= LABCARD_MAX)) {
	  return -ENODEV;}

	if(channel_busy[minor]) {
	  return -EBUSY;}
	channel_busy[minor]=1;

	if(minor == LABCARD_OUTPUT_CHANNEL_0) {
	  outb((char)0,(unsigned short)(baseI+4)); 
	  /* reset of D/A channel #1 low byte */
	
	  outb((char)0,(unsigned short)(baseI+5)); 
	  /* reset of D/A channel #1 high byte */}


	if(minor == LABCARD_OUTPUT_CHANNEL_1) {    
	  outb((char)0,(unsigned short)(baseI+6)); 
	  /* reset of D/A channel #2 low byte */
	
	  outb((char)0,(unsigned short)(baseI+7)); 
	  /* reset of D/A channel #2 high byte */}

	return 0;
}

static int labcard_read(struct inode * inode, struct file * file, char * buffer, int count)
{
        int minor = MINOR(inode->i_rdev);
	if(minor >= LABCARD_INPUT_CHANNEL) {
	  return -EINVAL;}

	put_fs_word((short) read_AD((char)minor),buffer);
	return 2;
}

static int labcard_write(struct inode * inode, struct file * file, char * buffer, int count)
{
        int minor = MINOR(inode->i_rdev);
	if((minor < LABCARD_INPUT_CHANNEL)||(minor >= LABCARD_MAX)) {
	  return -EINVAL;}
	
	if(write_DA((minor - LABCARD_INPUT_CHANNEL),get_fs_word(buffer))) {
	  return 2;}
	else {
	  return -EINVAL;}
 }

static void labcard_release(struct inode * inode, struct file * file)
{
	int minor = MINOR(inode->i_rdev);
	if((minor < LABCARD_MIN)||(minor >= LABCARD_MAX)) {
	  return /*-ENODEV*/;}
	channel_busy[minor]=0;

	if(minor == LABCARD_OUTPUT_CHANNEL_0) {
	  outb((char)0,(unsigned short)(baseI+4)); 
	  /* reset of D/A channel #1 low byte */
	
	  outb((char)0,(unsigned short)(baseI+5)); 
	  /* reset of D/A channel #1 high byte */}


	if(minor == LABCARD_OUTPUT_CHANNEL_1) {    
	  outb((char)0,(unsigned short)(baseI+6)); 
	  /* reset of D/A channel #2 low byte */
	
	  outb((char)0,(unsigned short)(baseI+7)); 
	  /* reset of D/A channel #2 high byte */}

	return;
}

static int labcard_ioctl(struct inode * inode, struct file * file, unsigned int cmd, unsigned long arg)
{
	int minor = MINOR(inode->i_rdev);

	switch(cmd)
	  {
	  case LC_GAIN: switch(arg) {
	                         case 0:
                                 case 1:
	                         case 2:
	                         case 3:
	                         case 4:
                                        outb((char)arg,(unsigned short)(baseI+9));
					break;
				      default: return -EINVAL;
				      }
	                break;
	  case LC_ADDRESS: switch(arg) {
	                         case 0x200:
	                         case 0x210:
	                         case 0x220:
	                         case 0x230:
	                         case 0x300:
	                         case 0x3F0:
	                                    baseI = arg;
					    break;
					  default: return -EINVAL;
					  }
	                   break;
	  }
	return 0;
}

static struct file_operations labcard_fops = {
        NULL,		/* seek */
	labcard_read,	/* read */
	labcard_write,  /* write */
	NULL,		/* readdir */
	NULL,		/* select */
	labcard_ioctl,	/* ioctl */
	NULL,		/* mmap */
        labcard_open,   /* open */
        labcard_release,/* release */
};

unsigned long labcard_init(unsigned long kmem_start)
{
	if (register_chrdev(LABCARD_MAJOR,"lc",&labcard_fops)){
		printk("unable to get major %d for labcard devices\n",
		       LABCARD_MAJOR);}
	else {
	  outb((char)(baseI+11),(unsigned short)0x01);       /* enable software trigger */
	  outb((char)0,(unsigned short)(baseI+4)); 
	  /* reset of D/A channel #1 low byte */
	
	  outb((char)0,(unsigned short)(baseI+5)); 
	  /* reset of D/A channel #1 high byte */

	  outb((char)0,(unsigned short)(baseI+6)); 
	  /* reset of D/A channel #2 low byte */
	
	  outb((char)0,(unsigned short)(baseI+7)); 
	  /* reset of D/A channel #2 high byte */

	  printk("labcard driver: labcard.c v.0.0.1 01/27/1995 (Ludo Koren) koren@vm.stuba.sk\n");
	}
	return kmem_start;
}
